@@ -1,7 +1,6 @@
-#!/usr/bin/python3
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   MenuLibre - Advanced fd.o Compliant Menu Editor
-#   Copyright (C) 2012-2021 Sean Davis <sean@bluesabre.org>
+#   Copyright (C) 2012-2024 Sean Davis <sean@bluesabre.org>
 #   Copyright (C) 2017-2018 OmegaPhil <OmegaPhil@startmail.com>
 #
 #   This program is free software: you can redistribute it and/or modify it
@@ -16,25 +15,28 @@
 #   You should have received a copy of the GNU General Public License along
 #   with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+import logging
 import os
 import re
 import subprocess
 
 import getpass
 import psutil
+import shutil
 
 from locale import gettext as _
 
-from gi.repository import GLib, Gdk
+import gi
+gi.require_version('Gdk', '3.0')
+from gi.repository import GLib, Gdk  # type: ignore
 
-import logging
 logger = logging.getLogger('menulibre')
 
 old_psutil_format = isinstance(psutil.Process.username, property)
 
 
 def enum(**enums):
-    """Add enumarations to Python."""
+    """Add enumerations to Python."""
     return type('Enum', (), enums)
 
 
@@ -65,7 +67,7 @@ MenuItemKeys = (
     ("Exec", str, True, (0,)),
     ("Path", str, False, (0,)),
     ("Terminal", bool, False, (0,)),
-    ("Actions", list, False, (0,)),
+    ("Actions", str, False, (0,)),
     ("MimeType", list, False, (0,)),
     ("Categories", list, False, (0,)),
     ("Implements", list, False, (0,)),
@@ -79,11 +81,11 @@ MenuItemKeys = (
 def getRelatedKeys(menu_item_type, key_only=False):
     if isinstance(menu_item_type, str):
         if menu_item_type == "Application":
-            menu_item_type = MenuItemTypes.APPLICATION
+            menu_item_type = MenuItemTypes.APPLICATION  # type: ignore
         elif menu_item_type == "Link":
-            menu_item_type = MenuItemTypes.LINK
+            menu_item_type = MenuItemTypes.LINK  # type: ignore
         elif menu_item_type == "Directory":
-            menu_item_type = MenuItemTypes.DIRECTORY
+            menu_item_type = MenuItemTypes.DIRECTORY  # type: ignore
 
     results = []
     for tup in MenuItemKeys:
@@ -135,7 +137,7 @@ def getProcessList():
     """Return a list of unique process names for the current user."""
     username = getpass.getuser()
     try:
-        pids = psutil.get_pid_list()
+        pids = psutil.get_pid_list()  # type: ignore
     except AttributeError:
         pids = psutil.pids()
     processes = []
@@ -153,11 +155,25 @@ def getProcessList():
     return processes
 
 
-def getBasename(filename):
+def getRelativeName(filename: str):
     if filename.endswith('.desktop'):
         basename = filename.split('/applications/', 1)[1]
     elif filename.endswith('.directory'):
         basename = filename.split('/desktop-directories/', 1)[1]
+    else:
+        basename = ""
+    return basename
+
+
+def getBasename(filename: str):
+    if filename.endswith('.desktop'):
+        basename = filename.split('/applications/', 1)[1]
+    elif filename.endswith('.directory'):
+        basename = filename.split('/desktop-directories/', 1)[1]
+        if basename.startswith("%s/" % getDefaultMenuName()):
+            basename = filename.split("%s/" % getDefaultMenuName(), 1)[1]
+    else:
+        basename = ""
     return basename
 
 
@@ -175,7 +191,14 @@ def getCurrentDesktop():
     return current_desktop
 
 
-def getDefaultMenuPrefix(): # noqa
+def getDefaultMenuName():
+    prefix = getDefaultMenuPrefix()
+    if prefix.endswith("-"):
+        return prefix[:-1]
+    return prefix
+
+
+def getDefaultMenuPrefix():  # noqa
     """Return the default menu prefix."""
     prefix = os.environ.get('XDG_MENU_PREFIX', '')
 
@@ -185,6 +208,9 @@ def getDefaultMenuPrefix(): # noqa
             prefix = 'cinnamon-'
         elif 'mate' in os.environ.get('DESKTOP_SESSION', ''):
             prefix = 'mate-'
+        # Somehow the XDG_MENU_PREFIX isn't exposed in Ubuntu Unity
+        elif 'unity' in os.environ.get('DESKTOP_SESSION', ''):
+            prefix = 'gnome-'
 
     if prefix == "":
         desktop = getCurrentDesktop()
@@ -219,7 +245,7 @@ def getMenuDiagnostics():
         diagnostics[k] = os.environ.get(k, "None")
 
     menu_dirs = [
-        getUserMenuPath()
+        getUserMenusDirectory()
     ]
     for path in GLib.get_system_config_dirs():
         menu_dirs.append(os.path.join(path, 'menus'))
@@ -238,16 +264,23 @@ def getMenuDiagnostics():
     return diagnostics
 
 
+def getItemSearchPaths():
+    search_paths = []
+    for path in GLib.get_system_data_dirs():
+        search_paths.append(os.path.join(path, 'applications'))
+    return search_paths
+
+
 def getItemPath(file_id):
     """Return the path to the system-installed .desktop file."""
-    for path in GLib.get_system_data_dirs():
-        file_path = os.path.join(path, 'applications', file_id)
+    for path in getItemSearchPaths():
+        file_path = os.path.join(path, file_id)
         if os.path.isfile(file_path):
             return file_path
     return None
 
 
-def getUserItemPath():
+def getUserApplicationsDirectory():
     """Return the path to the user applications directory."""
     item_dir = os.path.join(GLib.get_user_data_dir(), 'applications')
     if not os.path.isdir(item_dir):
@@ -255,16 +288,120 @@ def getUserItemPath():
     return item_dir
 
 
+def getUserItemPath(file_id):
+    """Return the path to the system-installed .desktop file."""
+    path = getUserApplicationsDirectory()
+    file_path = os.path.join(path, file_id)
+    if os.path.isfile(file_path):
+        return file_path
+    return None
+
+
+def getDirectorySearchPaths():
+    search_paths = []
+    for path in GLib.get_system_data_dirs():
+        search_paths.append(os.path.join(path, 'desktop-directories'))
+    return search_paths
+
+
 def getDirectoryPath(file_id):
     """Return the path to the system-installed .directory file."""
-    for path in GLib.get_system_data_dirs():
-        file_path = os.path.join(path, 'desktop-directories', file_id)
+    for path in getDirectorySearchPaths():
+        file_path = os.path.join(path, file_id)
         if os.path.isfile(file_path):
             return file_path
     return None
 
 
-def getUserDirectoryPath():
+def mapDesktopEnvironmentDirectories():
+    """
+    This feels wrong, but to make GMenu correctly handle desktop directories
+    in subdirectories, we need to bring them up to the top level.
+    """
+    menu = getDefaultMenuName()
+    if len(menu) == 0:
+        return
+
+    file_ids = []
+    for path in getDirectorySearchPaths():
+        if not os.path.exists(path):
+            continue
+        for filename in os.listdir(path):
+            if filename.endswith(".desktop"):
+                file_ids.append(filename)
+
+    de_paths = {}
+    for path in GLib.get_system_data_dirs():
+        de_path = os.path.join(path, menu, 'desktop-directories')
+        if not os.path.exists(de_path):
+            continue
+        for filename in os.listdir(de_path):
+            if filename not in file_ids:
+                de_paths[filename] = de_path
+
+    user_dir = getUserDirectoriesDirectory()
+    for filename, basedir in de_paths.items():
+        if filename in file_ids:
+            continue
+
+        target_dir = os.path.join(user_dir, menu)
+        try:
+            os.makedirs(target_dir)
+        except BaseException:
+            pass
+        target = os.path.join(target_dir, filename)
+        src = os.path.join(basedir, filename)
+        if not os.path.exists(target):
+            try:
+                logger.debug("copy %s to %s" % (src, target))
+                shutil.copy2(src, target)
+            except BaseException:
+                logger.warning("Failed to copy %s to %s" % (src, target))
+                continue
+
+        try:
+            symlink = os.path.join(user_dir, filename)
+            if os.path.exists(symlink):
+                continue
+            os.symlink(target, symlink)
+        except BaseException:
+            logger.warning("Failed to symlink %s to %s" % (src, target))
+
+
+def unmapDesktopEnvironmentDirectories():
+    """
+    This feels wrong, but to make GMenu correctly handle desktop directories
+    in subdirectories, we need to bring them up to the top level.
+    """
+    menu = getDefaultMenuName()
+    if len(menu) == 0:
+        return
+
+    user_dir = getUserDirectoriesDirectory()
+
+    de_dir = os.path.join(user_dir, menu)
+    if not os.path.exists(de_dir):
+        return
+
+    for filename in os.listdir(user_dir):
+        filename = os.path.join(user_dir, filename)
+
+        if not filename.endswith(".directory"):
+            continue
+        if not os.path.islink(filename):
+            continue
+
+        realpath = os.path.realpath(filename)
+        if not realpath.startswith(de_dir):
+            continue
+
+        try:
+            os.remove(filename)
+        except BaseException:
+            logger.warning("Failed to remove symlink %s" % filename)
+
+
+def getUserDirectoriesDirectory():
     """Return the path to the user desktop-directories directory."""
     menu_dir = os.path.join(GLib.get_user_data_dir(), 'desktop-directories')
     if not os.path.isdir(menu_dir):
@@ -272,7 +409,16 @@ def getUserDirectoryPath():
     return menu_dir
 
 
-def getUserMenuPath():
+def getUserDirectoryPath(file_id):
+    """Return the path to the system-installed .directory file."""
+    path = getUserDirectoriesDirectory()
+    file_path = os.path.join(path, file_id)
+    if os.path.isfile(file_path):
+        return file_path
+    return None
+
+
+def getUserMenusDirectory():
     """Return the path to the user menus directory."""
     menu_dir = os.path.join(GLib.get_user_config_dir(), 'menus')
     if not os.path.isdir(menu_dir):
@@ -283,14 +429,9 @@ def getUserMenuPath():
 def getUserLauncherPath(basename):
     """Return the user-installed path to a .desktop or .directory file."""
     if basename.endswith('.desktop'):
-        check_dir = "applications"
+        return getUserItemPath(basename)
     else:
-        check_dir = "desktop-directories"
-    path = os.path.join(GLib.get_user_data_dir(), check_dir)
-    filename = os.path.join(path, basename)
-    if os.path.isfile(filename):
-        return filename
-    return None
+        return getUserDirectoryPath(basename)
 
 
 def getSystemMenuPath(file_id):
@@ -305,15 +446,9 @@ def getSystemMenuPath(file_id):
 def getSystemLauncherPath(basename):
     """Return the system-installed path to a .desktop or .directory file."""
     if basename.endswith('.desktop'):
-        check_dir = "applications"
+        return getItemPath(basename)
     else:
-        check_dir = "desktop-directories"
-    for path in GLib.get_system_data_dirs():
-        path = os.path.join(path, check_dir)
-        filename = os.path.join(path, basename)
-        if os.path.isfile(filename):
-            return filename
-    return None
+        return getDirectoryPath(basename)
 
 
 def getDirectoryName(directory_str):  # noqa
@@ -493,18 +628,20 @@ def getSaveFilename(name, filename, item_type, force_update=False):  # noqa
 
         # Use the current filename as a base.
         else:
-            basename = getBasename(filename)
+            basename = getRelativeName(filename)
 
         # Split the basename into filename and extension.
         name, ext = os.path.splitext(basename)
 
         # Get the save location of the launcher base on type.
         if item_type == 'Application':
-            path = getUserItemPath()
+            path = getUserApplicationsDirectory()
             ext = '.desktop'
         elif item_type == 'Directory':
-            path = getUserDirectoryPath()
+            path = getUserDirectoriesDirectory()
             ext = '.directory'
+        else:
+            path = ""
 
         basedir = os.path.dirname(os.path.join(path, basename))
         if not os.path.exists(basedir):
@@ -575,7 +712,10 @@ def check_keypress(event, keys):  # noqa
             return False
     if 'Escape' in keys:
         keys[keys.index('Escape')] = 'escape'
-    if Gdk.keyval_name(event.get_keyval()[1]).lower() not in keys:
+    keyval_name = Gdk.keyval_name(event.get_keyval()[1])
+    if keyval_name is None:
+        return False
+    if keyval_name.lower() not in keys:
         return False
 
     return True
@@ -615,7 +755,7 @@ def find_program(program):
     if len(program) == 0:
         return None
 
-    params = list(GLib.shell_parse_argv(program)[1])
+    params = list(GLib.shell_parse_argv(program)[1])  # type: ignore
     executable = params[0]
 
     if os.path.exists(executable):
@@ -627,6 +767,7 @@ def find_program(program):
 
     return None
 
+
 def validate_desktop_file(desktop_file):  # noqa
     """Validate a known-bad desktop file in the same way GMenu/glib does, to
     give a user real information about why certain files are broken."""
@@ -669,7 +810,7 @@ def validate_desktop_file(desktop_file):  # noqa
 
     # Type validation
     try:
-        type_key = keyfile.get_string(start_group,
+        type_key = keyfile.get_string(start_group,  # type: ignore
                                       GLib.KEY_FILE_DESKTOP_KEY_TYPE)
     except GLib.Error:
         # Translators: This error is displayed when a required key is
@@ -695,7 +836,7 @@ def validate_desktop_file(desktop_file):  # noqa
     # is not executable, the entry may be ignored (not be used in menus, for
     # example)."
     try:
-        try_exec = keyfile.get_string(start_group,
+        try_exec = keyfile.get_string(start_group,  # type: ignore
                                       GLib.KEY_FILE_DESKTOP_KEY_TRY_EXEC)
     except GLib.Error:
         pass
@@ -709,10 +850,10 @@ def validate_desktop_file(desktop_file):  # noqa
 
     # Validating executable
     try:
-        exec_key = keyfile.get_string(start_group,
+        exec_key = keyfile.get_string(start_group,  # type: ignore
                                       GLib.KEY_FILE_DESKTOP_KEY_EXEC)
     except GLib.Error:
-        return False # LP: #1788814, Exec key is not required
+        return False  # LP: #1788814, Exec key is not required
 
     try:
         if find_program(exec_key) is None:
@@ -722,9 +863,9 @@ def validate_desktop_file(desktop_file):  # noqa
         return (_('%s program \'%s\' is not a valid shell command '
                   'according to GLib.shell_parse_argv, error: %s')
                 % ('Exec', exec_key, e))
-    
+
     if type_key == "Service":
-        return False # KDE services are not displayed in the menu
+        return False  # KDE services are not displayed in the menu
 
     # Translators: This error is displayed for a failing desktop file where
     # errors were detected but the file seems otherwise valid.
